# This tool verifies the clauses generated by SMS. Lines starting with "c\t" are ignored
#
# The first line gives the ordering of the r-subsets used for the symmetry breaking.
# The next lines until the line "clauses:" gives the translation from the solvers variable
# to a r-tuple in the following format "<var> <r-tuple>"
#
# All other lines after "clauses:" have the following format: clause;reason;info
# The additional info is optional
#
# We distinguish between 4 type of clauses:
# - for basis exchange property: b
# - symmetry breaking clauses: s
# - for encoding the number of disjoint rainbow bases: t
# - for encoding the girth: g


import argparse
import ast
import itertools
from sys import stdin


class IllegalStateError(RuntimeError):
    pass


parser = argparse.ArgumentParser(
    description='Checking if the clauses are correct given a file.')
parser.add_argument('--rank', "-r", type=int,
                    help='the rank of the matroid', required=True)
parser.add_argument('--inputFile', "-i", type=str,
                    help='the input file for checking')
parser.add_argument('--transversals', "-t", "-k", type=int)
parser.add_argument('--girth', "-g", type=int)
args = parser.parse_args()

rank = args.rank

inputFile = None
if args.inputFile:
    inputFile = open(args.inputFile, "r")
else:
    inputFile = stdin


def binomialCoeff(n, k):
    if k > n:
        return 0
    if k == 0 or k == n:
        return 1
    return binomialCoeff(n-1, k-1) + binomialCoeff(n-1, k)


nSubsets = binomialCoeff(rank ** 2, rank)
nElems = rank ** 2

threshold = 100  # some additional variables
# Maps each variable to the base representing it (or [] if not representing a base)
var2subset = [None for _ in range(nSubsets + threshold)]
orderOfSubsets = []  # Will store the order of r-subsets for the symmetry breaking


def checkBaseExchange(lits, info):
    element = int(info.strip())

    if lits[0] > 0 or lits[1] > 0:
        raise IllegalStateError("Error: first two must be negative")

    for i in range(2, len(lits)):
        if lits[i] < 0:
            raise IllegalStateError("Error: lits must be negative")

    basis1 = var2subset[abs(lits[0])]
    basis2 = var2subset[abs(lits[1])]
    # map all the other literals
    other = list(map(lambda x: var2subset[x], lits[2:]))

    if element in basis2:
        basis1, basis2 = basis2, basis1

    if element not in basis1 or element in basis2:
        raise IllegalStateError(
            "Error: element must be exactly in one")

    diff = [e for e in basis2 if e not in basis1]
    if len(diff) != len(lits) - 2:
        raise IllegalStateError("Error: number of literals does not match for basis exchange", len(
            diff), "vs", len(lits) - 2)

    for e in diff:
        newSubset = [e2 for e2 in basis1 if e2 != element] + [e]
        newSubset.sort()
        if newSubset not in other:
            raise IllegalStateError("Error: r-subset missing")


def checkGirth(lits, info):
    # Clause ensuring that elems is an independent set
    elems = ast.literal_eval(info)
    if len(elems) + 1 != args.girth or len(elems) != len(set(elems)):
        raise IllegalStateError("Error: number of elements must be girth - 1")

    subsets = list(map(lambda x: var2subset[x], lits))
    subsets.sort()

    # complement of elems
    otherElems = [e for e in range(rank ** 2) if e not in elems]
    subsets2 = list(map(lambda x: sorted(elems + list(x)),
                        itertools.combinations(otherElems, rank - len(elems))))
    subsets2.sort()
    if subsets != subsets2:
        raise IllegalStateError("Error: bases should be the same")


def checkSymmetry(lits, info):
    permutation = ast.literal_eval(info)
    if sorted(permutation) != list(range(rank**2)):
        raise IllegalStateError("Error: not a permutation")

    # Check if each row maps on another row
    for i in range(rank):
        row = list(range(rank*i, rank*(i + 1)))
        permutedRow = [permutation[e] for e in row]
        if max(permutedRow) - min(permutedRow) != rank - 1 or min(permutedRow) % rank != 0:
            raise IllegalStateError("Error: row does not map to another row")

    currentPos = 0  # literals checked up to the current position
    correct = False

    for i in range(len(orderOfSubsets)):
        subset = orderOfSubsets[i]
        permSubset = [permutation[e] for e in subset]
        permSubset.sort()

        #print(basis, permBasis)

        if subset == permSubset:
            continue

        if currentPos < len(lits) - 2:
            if lits[currentPos] < 0:
                if permSubset != var2subset[abs(lits[currentPos])]:
                    raise IllegalStateError("Error: wrong symmetry clause")
            else:
                if subset != var2subset[abs(lits[currentPos])]:
                    raise IllegalStateError("Error: wrong symmetry clause")

            currentPos = currentPos + 1
        else:  # last two elements
            if lits[len(lits) - 2] < 0 or lits[len(lits) - 1] > 0:
                raise IllegalStateError("Error: wrong symmetry clause")

            if var2subset[abs(lits[len(lits) - 2])] != subset:
                raise IllegalStateError("Error: wrong symmetry clause",)

            if var2subset[abs(lits[len(lits) - 1])] != permSubset:
                raise IllegalStateError("Error: wrong symmetry clause")
            correct = True
            break

    if not correct:
        raise IllegalStateError("Error: wrong symmetry clause")


def checkTransversals(lits):
    if len(lits) != 1:
        raise IllegalStateError("Error: should be a single literal")
    lit = lits[0]
    basis = var2subset[abs(lit)]

    # check whether "R_i" (row) or "D_i" (col)
    isRow = basis[0] + 1 == basis[1]

    if isRow:
        if basis != [e + basis[0] for e in range(rank)] or basis[0] % rank != 0:
            raise IllegalStateError("Error: not some R_i")

        if lit < 0:
            raise IllegalStateError("Error: must be positive")
    else:
        for i in range(rank - 1):
            if basis[i] + rank != basis[i + 1]:
                raise IllegalStateError("Error: not some D_i")

        jthRow = basis[0]  # which column

        if jthRow == rank - 1:
            if lit > 0:
                raise IllegalStateError("Error: last column must be negative")
        elif jthRow == 0:
            if lit < 0:
                raise IllegalStateError("Error: first column must be positive")
        else:
            if not args.transversals:
                raise IllegalStateError("Error: no transversal given")
            if jthRow < args.transversals and lit < 0:
                raise IllegalStateError("Error: should be positive")
            if jthRow >= args.transversals and lit > 0:
                raise IllegalStateError("Error: should be negative")


state = 0

for line in inputFile:
    if line.startswith("c\t"):  # comments
        continue

    if state == 0:  # first line is the order of bases for the symmetry breaking
        orderOfSubsets = ast.literal_eval(line)

        # first rank bases must be the D_1,..., D_r
        for i in range(rank):
            if orderOfSubsets[i] != [i + rank * j for j in range(rank)]:
                raise IllegalStateError(
                    "Error: first bases must be the columns")

        for subset in orderOfSubsets[rank:]:
            if subset != sorted(subset) or len(subset) != len(set(subset)) or len(subset) != rank:
                raise IllegalStateError("Not sorted or duplicates")

            # same as [i*r + 0, i*r + 1,..., i*r + r - 1] for some i in [r]
            if max(subset) - min(subset) == rank - 1 and min(subset) % rank == 0:
                raise IllegalStateError(
                    "Error: order of bases not allowed to contain some R_i")

        print("Checked ordering of bases")
        state = 1

    elif state == 1:  # Read translation from variables to subsets
        if line.startswith("clauses:"):
            state = 2

            subsets = [b for b in var2subset if b != None]
            subsets.sort()

            for i in range(len(subsets) - 1):
                if subsets[i] == subsets[i + 1]:
                    raise IllegalStateError("Error: no duplicates allowed")

            if len(subsets) != nSubsets:
                raise IllegalStateError("Error: bases missing")

            print("Checked translation of variables")
        else:
            v, subset = line.split()
            index = int(v)
            var2subset[index] = ast.literal_eval(subset)

            if len(var2subset[index]) != rank:
                raise IllegalStateError(
                    "Error: bases with wrong amount of elements")

            for i in range(rank):
                if type(var2subset[index][i]) != int or var2subset[index][i] < 0 or var2subset[index][i] >= rank ** 2:
                    raise IllegalStateError("Error: wrong type")

            if var2subset[index] != sorted(var2subset[index]) or len(var2subset[index]) != len(set(var2subset[index])):
                raise IllegalStateError("Error: should be ordered")

    elif state == 2:  # Check clauses
        splittedLine = line.split(";")
        if len(splittedLine) == 1:
            raise IllegalStateError("Error: every clause must have a reason")
        lits = list(map(int, splittedLine[0].split()))
        if lits[-1] != 0:
            raise IllegalStateError("Error: last \"literal\" must be zero")
        lits = lits[:len(lits) - 1]

        reason = splittedLine[1].strip()

        if reason == 'b':  # base exchange property
            checkBaseExchange(lits, splittedLine[2])

        elif reason == 't':
            checkTransversals(lits)

        elif reason == 's':  # minimizing the negation of the variables
            checkSymmetry(lits, splittedLine[2])

        elif splittedLine[1] == 'g':
            checkGirth(lits, splittedLine[2])
        else:
            raise IllegalStateError("Error: no valid reason")

if state != 2:
    raise IllegalStateError("Error: not a single clause added")

print("Everything okay")
